def load_variable_list(filename):
    # Load list of variables
    from numpy import loadtxt
    variables = loadtxt(filename, dtype=int).tolist()
    if not isinstance(variables, list):
        variables = [variables]
    return variables


def load_source_files(path):
    filelist = [ReadSourceFile(join(path, filename)) for filename in listdir_ord(path) if filename.count(".f90")]
    filedict = {}
    for f in filelist:
        filedict[f.filename] = f
    return filedict


def listdir_ord(path):
    from os import listdir
    file_list = listdir(path)
    file_list.sort()
    return file_list

def load_intrinsics():
    package_directory = dirname(abspath(__file__))
    # Hardcoded path to file.
    with open(package_directory + "/../AdditionalFiles/fortran_intrinsics") as f:
        return [re.search(r"(\w+)", l).groups(0)[0].lower() for l in f]

def load_vault(filename="vault.pkl"):
    try:
        return pickle.load(open(filename, "rb"))
    except IOError:
        return None


def load_list_of_files_to_keep_unmodified(filename=None):
    if filename is None:
        package_directory = dirname(abspath(__file__))
        filename = package_directory + "/../AdditionalFiles/list_of_files_to_keep_unmodified.txt"
    # Hardcoded path to file.
    if isfile(filename):
        print("Loading list of exceptions from %s\n\n" % filename)
        with open(filename) as f:
            return [l.strip() for l in f if l.strip() and l.strip()[0] != "#"]
    else:
        return None


def save_variable_list(list_of_variables, filename, vault):
    # Save list of variables into a txt file
    # Sort and keep only uniq values
    list_of_variables = list(sorted(set(list_of_variables)))
    line_format = "%5i    # %20s %25s %20s\n"
    with open(filename, "w") as output_file:
        # Add header
        output_file.write("#%5s     %20s %25s %20s\n%s\n" % ("ID", "variable", "routine", "module", 100 * "#"))
        # Add variables
        for var_id in list_of_variables:
            v = vault.get_variable_by_id(var_id)
            output_file.write(line_format % (v.id, v.name, v.routine.name, v.routine.module.name))
